#include "ETCallBackHandler.h"
#include "RootRecord.h"
//#include <boost/graph/graph_traits.hpp>
//#include <boost/graph/adjacency_list.hpp>
//#include <boost/graph/topological_sort.hpp>
#include "main.h"
#include "ETThreadLocalData.h"
#include <agent_util.h>
#include <java_crw_demo.h>
#include <vector>
#include <set>
#include <iostream>
#include <algorithm>
#include <GraphAlgorithms.h>
#include <assert.h>
#include <sstream>
using namespace std;
// using namespace boost;

static void cleanerStaticWrapper (jvmtiEnv *jvmti, JNIEnv* jni, void* data) {
  ETCallBackHandler* etCallBackHandler = (ETCallBackHandler*)data;
  etCallBackHandler->weakRefCleanerThreadStart(jni);	
}

ETCallBackHandler::ETCallBackHandler (jvmtiEnv *jvmti, JavaVM* vm, TraceOutputter* traceOutputter) {
  CallBackHandler::CallBackHandler(jvmti, traceOutputter);
  this->traceOutputter = traceOutputter;
  this->aJVMTI = jvmti;
  this->currentTime = 0;
  this->doDeathChunks = includeDeathChunks;

  if (this->doDeathChunks) {
    this->deathChunkFile.open("deathChunks");
  }

  jvmti->CreateRawMonitor("heapLock"         , &heapLock         );
  jvmti->CreateRawMonitor("vertexMapLock"    , &vertexMapLock    );
  jvmti->CreateRawMonitor("deadListLock"     , &deadListLock     );
  jvmti->CreateRawMonitor("clockLock"        , &clockLock        );
  jvmti->CreateRawMonitor("tagLock"          , &tagLock          );
  jvmti->CreateRawMonitor("tag2timeStampLock", &tag2timeStampLock);
  //jvmti->CreateRawMonitor("arrayShadowLock"  , &arrayShadowLock  );

  jvmti->CreateRawMonitor("weakRefMon"       , &weakRefMon       );

  weakRefCleaner = new WeakRefCleaner(jvmti, weakRefMon, this);
}

void ETCallBackHandler::lockOn (jrawMonitorID lock) {
  jvmtiError error = aJVMTI->RawMonitorEnter(lock);
  check_jvmti_error(aJVMTI, error, "ETCallBackHandler::lockOn: Cannot enter with raw monitor");
}

void ETCallBackHandler::unlockOn (jrawMonitorID lock) {
  jvmtiError error = aJVMTI->RawMonitorExit(lock);
  check_jvmti_error(aJVMTI, error, "ETCallBackHandler::unlockOn:Cannot exit with raw monitor");
}

static ETThreadLocalData *getThreadLocalData (jvmtiEnv *jvmti_env, jthread thread, jint tag) {
  ETThreadLocalData *tld;
  jvmtiError err = jvmti_env->GetThreadLocalStorage(thread, (void **)(&tld));
  check_jvmti_error(jvmti_env, err,
                    "ETCallBackHandler::getThreadLocalData: Cannot GetThreadLocalStorage");
  if (tld == NULL) {
    tld = new ETThreadLocalData(tag);
    err = jvmti_env->SetThreadLocalStorage(thread, tld);
    check_jvmti_error(jvmti_env, err,
                      "ETCallBackHandler::getThreadLocalData: Cannot SetThreadLocalStorage");
  }

  return tld;
}

void ETCallBackHandler::nativeMethodEntry (JNIEnv *jni, jclass klass,
                                           jint methodId, jobject receiver) {
  jthread currentThread;
	
  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI, err, "mearthAgent: nativeMethodEntry: could not get current thread");
	
  jlong newTime = incTime();

  long receiverTag = getTagAndTagIfNeeded(jni, receiver     );
  long threadTag   = getTagAndTagIfNeeded(jni, currentThread);

  ETThreadLocalData *tld = getThreadLocalData(aJVMTI, currentThread, threadTag);
  tld->incrementCallDepth(1);
  tld->noteEntry(tld->getCallDepth(), methodId);

  this->traceOutputter->doMethodEntry(newTime, methodId, receiverTag, threadTag);
}

void ETCallBackHandler::nativeMethodEntryStatic (JNIEnv *jni, jclass klass, jint methodId) {
  //TODO: do we need/want to distinguish better?
  this->nativeMethodEntry(jni, klass, methodId, 0);
}

void ETCallBackHandler::nativeMethodExit (JNIEnv *jni, jclass klass,
                                          jint methodId, jobject receiver) {
  jthread currentThread;

  this->nativeVarEvent(jni, klass, 1, receiver, 0, 0, 0, 0);

  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI, err, "mearthAgent::nativeMethodExit: could not get current thread");

  jlong newTime = incTime();

  long receiverTag = getTagAndTagIfNeeded(jni, receiver     );
  long threadTag   = getTagAndTagIfNeeded(jni, currentThread);

  ETThreadLocalData *tld = getThreadLocalData(aJVMTI, currentThread, threadTag);
  tld->noteExit(tld->getCallDepth(), methodId);
  tld->incrementCallDepth(-1);

  this->traceOutputter->doMethodExit(newTime, methodId, receiverTag, threadTag);
}

void ETCallBackHandler::nativeMethodException (JNIEnv *jni, jclass klass,
                                               jobject exception, jint methodId, jobject receiver) {
  jthread currentThread;

  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI, err, "mearthAgent::nativeMethodException: could not get current thread");

  long threadTag   = getTagAndTagIfNeeded(jni, currentThread);

  ETThreadLocalData *tld = getThreadLocalData(aJVMTI, currentThread, threadTag);
  tld->noteExit(tld->getCallDepth(), methodId);
  tld->incrementCallDepth(-1);

  //TODO: Note exception thrown, and the exception, in the trace

  /*
  this->nativeVarEvent(jni, klass, 1, receiver, 0, 0, 0, 0);

  jlong newTime = incTime();

  long receiverTag = getTagAndTagIfNeeded(jni, receiver     );

  this->traceOutputter->doMethodExit(newTime, methodId, receiverTag, threadTag);
  */
}

void ETCallBackHandler::nativeMethodExceptionStatic (JNIEnv *jni, jclass klass,
                                                     jobject exception, jint methodId) {
  jthread currentThread;

  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI, err, "mearthAgent::nativeMethodException: could not get current thread");

  long threadTag   = getTagAndTagIfNeeded(jni, currentThread);

  ETThreadLocalData *tld = getThreadLocalData(aJVMTI, currentThread, threadTag);
  tld->noteExit(tld->getCallDepth(), methodId);
  tld->incrementCallDepth(-1);

  //TODO: Note exception thrown, and the exception, in the trace

  //this->nativeVarEvent(jni, klass, 1, receiver, 0, 0, 0, 0);

  /*
  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI, err, "mearthAgent::nativeMethodExceptionStatic: could not get current thread");

  jlong newTime = incTime();

  long threadTag = getTagAndTagIfNeeded(jni, currentThread);

  this->traceOutputter->doMethodExit(newTime, methodId, 0, threadTag);
  */
}

bool vmDead () {
  return false;	
}

ETCallBackHandler::~ETCallBackHandler () { }

jlong ETCallBackHandler::tagObject (jvmtiEnv *jvmti, jobject o) {
  static jlong nextTag = 1; //The next object tag to use, we skip 0 since we use that to identify null
  jlong thisTag;

  if (nextTag == 0) {
    verbose_println("!! tagObject: nextTag is 0.");
  }

  //EBM: This should be done under a lock and should check first that the
  //tag is not yet set. Even with the lock, it is best to use sync_fetch_and_add.
  lockOn(tagLock);
  jvmtiError err = jvmti->GetTag(o, &thisTag);
  check_jvmti_error(this->aJVMTI, err, "ETCallBackHandler::tagObject: Could not get tag.");
  if (thisTag == 0) {
    err = jvmti->SetTag(o, (thisTag = __sync_fetch_and_add(&nextTag, 1)));
    check_jvmti_error(this->aJVMTI, err, "ETCallBackHandler::tagObject: Could not set tag.");
  }
  unlockOn(tagLock);

#ifdef ET_DEBUG
  {
    static set<long> tagsUsed;
    assert(tagsUsed.count(thisTag) == 0);
    tagsUsed.insert(thisTag);

    //Check that this tag is used for only one object.
    //It is only possible to call GetObjectsWithTags
    //in the live phase, so we cannot do this check
    //if we are in another phase.
    jvmtiPhase phase;
    err = jvmti->GetPhase(&phase);
    check_jvmti_error(jvmti, err, "ETCallBackHandler::tagObject: could not get phase.");
    if (phase == JVMTI_PHASE_LIVE) {
      jlong tags[1];
      tags[0] = thisTag;
      jint getTagCount = 0;

      err = jvmti->GetObjectsWithTags(1, tags, &getTagCount, NULL, NULL);
      check_jvmti_error(this->aJVMTI, err,
                        "ETCallBackHandler::tagObject: Could not get objects with tag: " + thisTag);
      assert(getTagCount == 1); 
    }		
  }
#endif	
  return thisTag;
}

void ETCallBackHandler::addObjectToHeap (jlong tag) {
  //lockOn(vertexMapLock);
  //obj2vertex[tag] = add_vertex(heap);
  //vertex2obj[obj2vertex[tag]] = tag;
  //unlockOn(vertexMapLock);
  //unlockOn(vertexMapLock);
}

void ETCallBackHandler::nativeNewObj (JNIEnv *jni, jclass klass,
                                      jobject thread, jobject o, bool isWeak) {
  jlong obj_size = 0;
  char* classSig;
  jthread currentThread;
  //RootEnumerator rootEnum(this->aJVMTI);
  verbose_println("ETCallBackHandler::nativeNewObj called. ");
  if (vmDead()) {
    return;	
  }

  jlong tag = tagObject(this->aJVMTI, o);

  jvmtiError err = this->aJVMTI->GetObjectSize(o, &obj_size);
  check_jvmti_error(this->aJVMTI, err, "ETCallBackHandler::nativeNewObj: could not get object size.");

  jclass classOfo = jni->GetObjectClass(o);
  err = this->aJVMTI->GetClassSignature(classOfo, &classSig, NULL);
  check_jvmti_error(this->aJVMTI, err,
                    "ETCallBackhandler::nativeNewObj: could not get class signature.");

  err = aJVMTI->GetCurrentThread(&currentThread);
  jlong threadTag = getTagAndTagIfNeeded(jni, currentThread);

  this->traceOutputter->doAllocation(currentTime, tag, obj_size, classSig, threadTag);

  timeStampTag(tag);
  addObjectToHeap(tag);

  if (isWeak) {
    jweak ref = jni->NewWeakGlobalRef(o);
    this->weakRefCleaner->addWeakRef(aJVMTI, ref);
  }

#ifdef DEBUG_MULTINODE
  if (strcmp(classSig,  "LtestPrograms/MultiNode;") == 0) {
    cerr << "NewObj: Added thing to multiNodes: " << hex << " 0x" << tag << endl;
    multiNodes.push_back(tag);		
  }
#endif 

  aJVMTI->Deallocate((unsigned char*)classSig);

  /*
    vector<long> rootList = rootEnum.roots();

    for(vector<long>::iterator it = rootList.begin();
    it != rootList.end();
    it++)
    {
    this->traceOutputter->doRoot(currentTime, *it);	
    }*/	

}

void ETCallBackHandler::nativeNewObj (JNIEnv *jni, jclass klass,
                                      jobject thread, jint classId, jint allocSite, bool isWeak) {
  //TODO: fill this out as we add the functionality
  //Note: needs to use nextTag, in another routine, eventually ...
  static long alternateTag = -1;

  long threadTag = getTagAndTagIfNeeded(jni, thread);
  ETThreadLocalData *tld = getThreadLocalData(aJVMTI, thread, threadTag);
  long newTag = __sync_fetch_and_add(&alternateTag, -1);
  long timestamp = this->currentTime;

  tld->newObject(allocSite, newTag, timestamp);
}

void ETCallBackHandler::nativeNewArray (JNIEnv *jni, jclass klass, jobject thread, jobject o) {
  //TODO: update when we switch to new allocation tracking scheme
  this->nativeNewObj(jni, klass, thread, o, false);
}

jlong ETCallBackHandler::getTagAndTagIfNeeded (JNIEnv* jni, jobject o) {
  jlong tag;
  jvmtiError err;
	
  if (jni->IsSameObject(o, NULL)) {
    //if the object is null, return 0 as its tag
    return 0;
  }

  err = this->aJVMTI->GetTag(o, &tag);
  check_jvmti_error(this->aJVMTI, err,
                    "ETCallBackHander::getTagAndTagIfNeeded: Could not get tag from object");

  if (tag == 0) {
    tag = tagObject(this->aJVMTI, o);
    if (tag == 0) {
      verbose_println("!! ETCallBackHandler::GetTagAndTagIfNeeded: got a zero tag.");
    }
    addObjectToHeap(tag);
    //TODO: Do I also want to time stamp this newly discovered object here?
    timeStampTag(tag);
  }

  return tag;		
}

void ETCallBackHandler::timeStampTag (jlong tag) {
  //Time stamping null doesn't make sense, ignore it.
  //Maybe we should tell the caller not to do this?
  if (tag != 0) {
    lockOn(tag2timeStampLock);
    //tag2TimeStamp.insert(pair<long,long>(tag, this->currentTime));
    tag2TimeStamp[tag] = this->currentTime;
    unlockOn(tag2timeStampLock);
  }
}

void ETCallBackHandler::ObjectFree (jvmtiEnv *jvmti, jlong tag) {	
  inDeathList = true;
#ifdef ET_DEBUG
  { //Check that we don't get ObjectFree calls twice for the same object.
    static set<long> seenObjects;

    //lockOn(tag2timeStampLock);
    cerr << seenObjects.count(tag) << " Object ID: 0x" << tag
         << "Time stamp: " << tag2TimeStamp[tag] << endl ;
    //unlockOn(tag2timeStampLock);

    assert(seenObjects.count(tag) == 0);
    seenObjects.insert(tag);
  }
#endif

#ifdef DEBUG_MULTINODE
  if (isMultiNode(tag)) {
    cerr << "ObjectFree: MultiNode Free " << hex << " 0x" << tag << endl;
  }
#endif
  //lockOn(deadListLock);
  if (tag == 0) {
    // cerr << "ObjectFree: Got Zero Tag" << endl;
  } else {
    //assert(count(deadList.begin(), deadList.end(), tag) == 0);
    deadList.push_back(tag);
    //assert(count(deadList.begin(), deadList.end(), tag) == 1);
  }

  //unlockOn(deadListLock);
  //unlock
}

void ETCallBackHandler::nativePointerMovedOff (JNIEnv *jni, jclass klass, jobject oldTarget,
                                               jobject origin, jobject newTarget) {
#ifdef THREAD_TAG_UPDATES
  jthread currentThread;
  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI, err, "mearthAgent: nativePointerMovedOff: could not get current thread");
#endif

  jlong oldTargetTag = getTagAndTagIfNeeded(jni, oldTarget);
  //A null origin indicates this is an update of a static field	
  jlong originTag    = getTagAndTagIfNeeded(jni, origin   );
  jlong newTargetTag = getTagAndTagIfNeeded(jni, newTarget);
	
#ifdef THREAD_TAG_UPDATES
  this->traceOutputter->doPointerUpdate(currentTime, oldTargetTag, originTag, newTargetTag,
                                        getTagAndTagIfNeeded(jni, currentThread));
#else  
  this->traceOutputter->doPointerUpdate(currentTime, oldTargetTag, originTag, newTargetTag);
#endif

  timeStampTag(oldTargetTag);
  timeStampTag(originTag);
  updateHeap(oldTargetTag, originTag, newTargetTag);
}


/*A pointer update, based only on tags
  Used by the WeakRefCleaner, who isn't so sure about objects */
void ETCallBackHandler::syntheticPointerMovedOff(JNIEnv *jni, jclass klass, jlong oldTargetTag, jlong originTag, jlong newTargetTag){
     
  timeStampTag(oldTargetTag);
  timeStampTag(originTag);
  updateHeap(oldTargetTag, originTag, newTargetTag);


#ifdef THREAD_TAG_UPDATES
  jthread currentThread;
  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI, err, "mearthAgent: nativePointerMovedOff: could not get current thread");
#endif

	
#ifdef THREAD_TAG_UPDATES
  this->traceOutputter->doPointerUpdate(currentTime, oldTargetTag, originTag, newTargetTag,
                                        getTagAndTagIfNeeded(jni, currentThread));
#else  
  this->traceOutputter->doPointerUpdate(currentTime, oldTargetTag, originTag, newTargetTag);
#endif


}


void ETCallBackHandler::nativeUninitPutfield (JNIEnv *jni, jclass klass, jobject newTarget,
                                              jint allocSite, jint fieldId) {
  jlong newTargetTag = getTagAndTagIfNeeded(jni, newTarget);
  timeStampTag(newTargetTag);
}

void ETCallBackHandler::nativeInitOf (JNIEnv *jni, jclass klass, jobject thread, jint allocSite) {
  long threadTag = getTagAndTagIfNeeded(jni, thread);
  ETThreadLocalData *tld = getThreadLocalData(aJVMTI, thread, threadTag);
  tld->initOf(allocSite);
}

void ETCallBackHandler::updateHeap (jlong oldTargetTag, jlong originTag, jlong newTargetTag) {
  lockOn(heapLock);
  if (oldTargetTag != 0) {
    heap.removeEdge(originTag,oldTargetTag);
  }

  if (newTargetTag != 0) {
    heap.addEdge(originTag,newTargetTag);
  }
  unlockOn(heapLock);
}

void ETCallBackHandler::nativeVarEvent (JNIEnv *jni, jclass klass, jint numObjs,
                                        jobject o0, jobject o1, jobject o2, jobject o3, jobject o4) {
  /*
    const char* c_className;
    const char* c_method;
    c_className = jni->GetStringUTFChars(className, NULL);
    c_method    = jni->GetStringUTFChars(method,    NULL);
    cout << "VAREVENT " << c_className << "/" << c_method << endl;
    jni->ReleaseStringUTFChars(className, c_className);
    jni->ReleaseStringUTFChars(method, c_method);
    delete(cpp_className);
    delete(cpp_method);
  */

  jlong objTag;
  jthread currentThread;

  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI,err, "ETCallBackHandler::nativeVarEvent: could not get current thread.");

  jlong threadTag = getTagAndTagIfNeeded(jni, currentThread);

  //EBM: does it matter that currentTime might change as we go?
  //The lack of breaks here is deliberate, we want the fall throgh behavior
  switch (numObjs) {
  case 5:
    objTag = getTagAndTagIfNeeded(jni, o4);
    timeStampTag(objTag);
    this->traceOutputter->doVarEvent(currentTime, objTag, threadTag);
  case 4:
    objTag = getTagAndTagIfNeeded(jni, o3);
    timeStampTag(objTag);
    this->traceOutputter->doVarEvent(currentTime, objTag, threadTag);
  case 3: 
    objTag = getTagAndTagIfNeeded(jni, o2);
    timeStampTag(objTag);
    this->traceOutputter->doVarEvent(currentTime, objTag, threadTag);
  case 2:
    objTag = getTagAndTagIfNeeded(jni, o1);
    timeStampTag(objTag);
    this->traceOutputter->doVarEvent(currentTime, objTag, threadTag);
  case 1:
    objTag = getTagAndTagIfNeeded(jni, o0);
    timeStampTag(objTag);
    this->traceOutputter->doVarEvent(currentTime, objTag, threadTag);
    break;
  default:
    //cerr << "ETCallBackHandler::nativeVarEvent: Unxpected number of objects " << numObjs
    //     << ", valid values are 1-5." << endl;
    exit(-1);
  }
}

//Caller must hold tag2TimeStamp lock...
bool ETCallBackHandler::diedBefore (unsigned long obj1, unsigned long obj2) {
  assert(tag2TimeStamp.count(obj1) != 0);
  assert(tag2TimeStamp.count(obj2) != 0);
  return tag2TimeStamp[obj1] < tag2TimeStamp[obj2];
}

//EBM: seems expensive ...
bool ETCallBackHandler::isUnreachable (unsigned long obj) {
  for (unsigned long i = 0; i < deadList.size(); i++) {
    if (deadList[i] == obj) {
      return true;	
    }	
  }
  return false;	
}

//Return the union of sets a and b.
static set<long> setUnion (const set<long> &a, const set<long> &b) {
  set<long> unionSet;

  for (set<long>::iterator it = a.begin(); it != a.end(); it++) {
    unionSet.insert(*it);
  }

  for (set<long>::iterator it = b.begin(); it != b.end(); it++) {
    unionSet.insert(*it);
  }

  return unionSet;
}

//Caller must hold locks on heap, tag2timeStamp
void ETCallBackHandler::computeObjectDeathTimes () {
  unsigned long lastTime = -1L; //This assumes two's complement to find max long.

  //stack<unsigned long int> unreachableStack;
  vector<unsigned long> unreachableStack; //Need own local copy of unreachable stack
  //sort unreachable objects from the earliest timestamp to the latest

  set<long> *completed = new set<long>();

  DiedBeforeComparator comp(*this);
  std::sort(deadList.begin(), deadList.end(), comp);//Send a function

  for (vector<unsigned long>::iterator it = deadList.begin(); it < deadList.end(); it++) {
    unreachableStack.push_back(*it);
  }

  while (!unreachableStack.empty()) {
    unsigned int obj = unreachableStack.back();
    unreachableStack.pop_back();

    pair<set<long>::iterator,bool> result = completed->insert(obj);
    if (result.second == false) continue;

    //We should have a time stamp for everything here...
    assert(tag2TimeStamp.count(obj) != 0);
    unsigned long  objTime = tag2TimeStamp[obj];

    if (objTime <= lastTime) {
      lastTime = objTime;

      //graph_traits < adjacency_list <hash_setS, vecS> >::adjacency_iterator neighborIt, lastNeighbor;
      //tie(neighborIt, lastNeighbor) = adjacent_vertices( obj2vertex[obj], heap);
      vector<long>::iterator neighborIt;

      pair<vector<long>::iterator, vector<long>::iterator  > beginEnd;

      beginEnd  = heap.neighbors(obj);

      for (neighborIt = beginEnd.first; neighborIt != beginEnd.second; ++neighborIt) {
        unsigned long  neighbor = *neighborIt;

        if (tag2TimeStamp.count(neighbor) == 0) {
          cerr << "Neighbor not found: 0x" << hex << neighbor << endl;
        }
        assert(tag2TimeStamp.count(neighbor) != 0);
        unsigned long neighborTime = tag2TimeStamp[neighbor];

        if (isUnreachable(neighbor) && neighborTime < lastTime) {
          tag2TimeStamp[ neighbor ] = lastTime;
          unreachableStack.push_back( neighbor );	
        }
      }

    }	

  }
}

void ETCallBackHandler::cleanUpLocalVarTable (jvmtiEnv* jvmti,
                                              jvmtiLocalVariableEntry *localVarTable,
                                              jint entryCount) {
  int i;

  for (i = 0; i < entryCount; i++) {
    jvmti->Deallocate((unsigned char*)localVarTable[i].name);
    jvmti->Deallocate((unsigned char*)localVarTable[i].signature);
    jvmti->Deallocate((unsigned char*)localVarTable[i].generic_signature);	
  }

  jvmti->Deallocate((unsigned char*)localVarTable);
}

//Time stamp the object o, and return its tag
jlong ETCallBackHandler::timeStampObject (jvmtiEnv *jvmti, jobject o) {
  jlong tag;
  jvmtiError err;

  err = jvmti->GetTag(o, &tag);
  check_jvmti_error(jvmti, err, "timeStampObject: Could not get tag.");

  timeStampTag(tag);

  return tag;
}

void ETCallBackHandler::nativeArrayStore (JNIEnv *jni, jclass klass,
                                          jobject array, jint index, jobject storee) {
#ifdef THREAD_TAG_UPDATES
  jthread currentThread;
  jvmtiError err = aJVMTI->GetCurrentThread(&currentThread);
  check_jvmti_error(aJVMTI, err, "ETCallBackHandler::nativeArrayStore: could not get current thread");
#endif

  jlong arrayTag  = getTagAndTagIfNeeded(jni, array );
  jlong storeeTag = getTagAndTagIfNeeded(jni, storee);

  //EBM: could there be a race condition here?
  //I really hope this cast is okay
  jobject oldTarget = jni->GetObjectArrayElement((jobjectArray)array, index);
  jlong oldTargetTag = getTagAndTagIfNeeded(jni, oldTarget);

  updateHeap(oldTargetTag, arrayTag, storeeTag);

#ifdef THREAD_TAG_UPDATES
  this->traceOutputter->doPointerUpdate(currentTime, oldTargetTag, arrayTag, storeeTag,
                                        getTagAndTagIfNeeded(jni,currentThread));
#else  
  this->traceOutputter->doPointerUpdate(currentTime, oldTargetTag, originTag, storeeTag);
#endif
}

void ETCallBackHandler::scanStack (jvmtiEnv *jvmti, int frame, JNIEnv *jni, jthread thread) {
  unsigned int framesToScan = 2;
  jobject obj;
  jint val;
  jvmtiError err;
  int count_slots = 0;
  int count_obj_slots = 0;
  jlong tag;

  // -- Acquire lock once for the whole stack scan
  lockOn(tag2timeStampLock);

  // int frame = 3;
  // for(unsigned int frame = 0; frame < framesToScan; frame++){

  for (unsigned int slot = 0; ; slot++) {
    err = jvmti->GetLocalObject(thread, frame, slot, &obj);
    // err = jvmti->GetLocalInt(thread, frame, slot, &val);
    //Scanning a slot where there is a double/the second half of a double appears to produce a
    //null object instead of producing a INVALID_SLOT error.
    if (err == JVMTI_ERROR_TYPE_MISMATCH) {
      // count_slots++;
      continue; //skip to the next slot
    } else if (err == JVMTI_ERROR_OPAQUE_FRAME) {
      break; //Can't scan this frame, skip to the next one.
    } else if (err == JVMTI_ERROR_NO_MORE_FRAMES) {
      break; // return; //Out of frames to scan
    } else if (err == JVMTI_ERROR_INVALID_SLOT) {
      break;//Looked at all the slots in this frame, go to next one
    } else {
      // check_jvmti_error(jvmti, err, "StackScanner::simpleScanStack: Could not get local object");
      // NOTFORCHECKIN
      if (!jni->IsSameObject(obj, NULL)) {
        // count_slots++;
        // count_obj_slots++;
        //tag2TimeStamp.insert(pair<long,long>(tag, this->currentTime));
        err = jvmti->GetTag(obj, &tag);
        tag2TimeStamp[tag] = this->currentTime;
        traceOutputter->doVarEvent(currentTime, tag,0);
        // timeStampObject(jvmti,obj);
      }

      // -- Not needed?
      // jni->DeleteLocalRef(obj); 
    }
  }
  // }

  unlockOn(tag2timeStampLock);
}

void ETCallBackHandler::GCStart (jvmtiEnv *jvmti) {
  verbose_println("Start GC...");
}

void ETCallBackHandler::GCEnd (jvmtiEnv *jvmti) {
  //EBM: I am concerned about running the weak ref processor concurrently
  //with other stuff ... will need to explore and think about it.
  jvmti->RawMonitorEnter(weakRefMon);
  jvmti->RawMonitorNotify(weakRefMon);
  jvmti->RawMonitorExit(weakRefMon);

  verbose_println("End GC");
}

void ETCallBackHandler::shadowGCEnd (jvmtiEnv *jvmti) {
  stringstream verbose_output;

  verbose_output << "...end GC -- got " << deadList.size() << " dead objects.";

  verbose_println(verbose_output.str());
  //We put these in a list to avoid holding a lock while entering the outputter.
  vector<pair<long,long> > deadObjsWithTime;
  
  //lockOn(deadListLock);
  //Slightly different computations, they
  // are side effect full and should not both be done.
  if (doDeathChunks) {
    //      lockOn(heapLock);
    //    lockOn(tag2timeStampLock);
    outPutDeathChunks(computeDeathChunks(heap, deadList, tag2TimeStamp),
                      deathChunkFile, tag2TimeStamp);
    //  unlockOn(tag2timeStampLock);
    //  unlockOn(heapLock);
  } else {	
    // lockOn(heapLock);
    //  lockOn(tag2timeStampLock);
    computeObjectDeathTimes();
    // unlockOn(tag2timeStampLock);
    // unlockOn(heapLock);
  }

  verbose_println("Output death records");

  // lockOn(heapLock);
  // lockOn(tag2timeStampLock);
  for (vector<unsigned long>::iterator it = deadList.begin(); it < deadList.end(); it++) {

    //traceOutputter->doDeath(*it, tag2TimeStamp[*it]);a
    //If we are creating new tags here, something has gone wrong...
    assert(tag2TimeStamp.count(*it) != 0);
    deadObjsWithTime.push_back(pair<long,long>(*it, tag2TimeStamp[*it]));

    heap.removeNode(*it);

    tag2TimeStamp.erase(*it);
  }
  // unlockOn(tag2timeStampLock);
  // unlockOn(heapLock);

  deadList.clear();
  // unlockOn(deadListLock);

  for (vector<pair<long,long> >::iterator it = deadObjsWithTime.begin();
       it != deadObjsWithTime.end();
       it++) {
    traceOutputter->doDeath(it->first, it->second);
  }

  traceOutputter->doGCEnd();
}

void ETCallBackHandler::VMInit (jvmtiEnv *jvmti, JNIEnv *jni, jthread thread) {

  //Start the cleaner thread
  //is there a better way to do this?
  jclass threadClass = jni->FindClass("java/lang/Thread");

  //Run agent thread needs a jthread, for some reason.
  //But I don't want to call its constructor,
  //Because I do not want it to show up in my trace.
  //AllocObject allocates one, but does not call the constructor.
  jthread weakCleanerThread = jni->AllocObject(threadClass);

  jvmtiError err = jvmti->RunAgentThread(weakCleanerThread, cleanerStaticWrapper, this,
                                         JVMTI_THREAD_NORM_PRIORITY);
  check_jvmti_error(jvmti, err,
                    "ETCallBackHandler::ETCallBackHandler: Could not start WeakRefCleaner thread.");
}

static jint scanLiveObjectsCBStaticWrapper (jvmtiHeapReferenceKind reference_kind,
                                            const jvmtiHeapReferenceInfo* reference_info,
                                            jlong class_tag,
                                            jlong referrer_class_tag,
                                            jlong size,
                                            jlong* tag_ptr,
                                            jlong* referrer_tag_ptr,
                                            jint length,
                                            void* user_data) {

  ETCallBackHandler* cbHandler = (ETCallBackHandler*)user_data;
  cbHandler->scanLiveObject(tag_ptr);

  if (referrer_tag_ptr != NULL) {
    cbHandler->scanLiveObject(referrer_tag_ptr);
  }

  return JVMTI_VISIT_OBJECTS;
}

static jint scanStringPrimitiveCBStaticWrapper (jlong class_tag, 
                                                jlong size, 
                                                jlong* tag_ptr, 
                                                const jchar* value, 
                                                jint value_length, 
                                                void* user_data) {
#ifdef ET_DEBUG
  cerr << "scanStringPrimitiveCBStaticWrapper called. Tag: " << *tag_ptr << endl;
#endif

  ETCallBackHandler* cbHandler = (ETCallBackHandler*)user_data;
  return cbHandler->scanLiveObject(tag_ptr);

}

static jint scanArrayPrimitiveCBStaticWrapper (jlong class_tag, 
                                               jlong size, 
                                               jlong* tag_ptr, 
                                               jint element_count, 
                                               jvmtiPrimitiveType element_type, 
                                               const void* elements, 
                                               void* user_data) {

#ifdef ET_DEBUG
  cerr << "scanArrayPrimitiveCBStaticWrapper caled. Tag: " << *tag_ptr << endl;
#endif

  ETCallBackHandler* cbHandler = (ETCallBackHandler*)user_data;
  return cbHandler->scanLiveObject(tag_ptr);
}

//Follow References will fill this set.
//Because there is no good way to get a return
//from it due to the JVMTI api, we will have to 
//have it fill this and remember to clear it.
//in fakeAGC
static set<long> liveObjects;
jint ETCallBackHandler::scanLiveObject (jlong* tag_ptr) {

  if (*tag_ptr != 0) {
    liveObjects.insert(*tag_ptr);
  }

  return JVMTI_VISIT_OBJECTS;
}

void ETCallBackHandler::VMDeath (jvmtiEnv *jvmti) {
  fakeAGC(jvmti);
  //traceOutputter->flush();
  if (doDeathChunks) {
    deathChunkFile.close();
  }

  //Delete global ref to Reference? Need a jni env...

}

void ETCallBackHandler::fakeAGC (jvmtiEnv *jvmti) {
  jvmtiError err;
  jvmtiHeapCallbacks heapCbs;
  vector<long> objsToFree;

  lockOn(deadListLock);
  lockOn(heapLock);
  lockOn(tag2timeStampLock);
  //We need to deal with whatever objects may be left in the heap.
  //First, we will iterate over live ones only, and time stamp them.
  memset (&heapCbs, 0, sizeof(heapCbs) );

  //Essentially the problem this function solves is that there may be dead objects
  //still hanging around in the heap, but we will never see the deallocation for them,
  //because the VM is dead, and we will get no more events.
  //So, first we scan all the live objects and time stamp them,
  //Then scan all objects, concluding that ones that do not have the current time stamp are dead.
  //Then we do an artificial GC.

  heapCbs.heap_reference_callback = scanLiveObjectsCBStaticWrapper;
  heapCbs.string_primitive_value_callback = scanStringPrimitiveCBStaticWrapper;
  heapCbs.array_primitive_value_callback = scanArrayPrimitiveCBStaticWrapper;
  //heapCbs.heap_iteration_callback = scanAllObjectsCBStaticWrapper;

  err = jvmti->FollowReferences(0, NULL, NULL, &heapCbs, (void*)this);
  check_jvmti_error(jvmti, err, "ETCallBackHandler::VMDeath could not follow references");

  //err = jvmti->IterateThroughHeap(0, NULL, &heapCbs, this);
  //Now go over all objects, to see who is dead*/
  //need to aquire a lock here?
  //lockOn(tag2timeStampLock);
  for (map<unsigned long, unsigned long>::iterator it = tag2TimeStamp.begin();
       it != tag2TimeStamp.end();
       it++) {
    if (liveObjects.count((*it).first) == 0) { //object is not live
      objsToFree.push_back((*it).first);
    }	
  } 
  //unlockOn(tag2timeStampLock);

  liveObjects.clear();

  for (vector<long>::iterator it = objsToFree.begin(); it != objsToFree.end(); it++) {
    ObjectFree(jvmti,*it);
  }
  this->shadowGCEnd(jvmti);

  unlockOn(deadListLock);
  unlockOn(heapLock);
  unlockOn(tag2timeStampLock);
}

void ETCallBackHandler::weakRefCleanerThreadStart (JNIEnv* jni) {
  weakRefCleaner->cleanerThreadStart(aJVMTI, jni, heap);
}

#ifdef DEBUG_MULTINODE
bool ETCallBackHandler::isMultiNode (unsigned long tag) {

  for (vector<unsigned long>::iterator it = multiNodes.begin(); it != multiNodes.end(); it++) {
    if (*it == tag) {
      return true;
    }
  }

  return false;
}
#endif

// Local Variables:
// mode:C++
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
