import java.lang.ref.*;

public class ElephantTracks{
	
  /**
   * indicates whether things are set up enough to start producing trace output
   */
  private static int engaged = 0;

  /**
   * native method to indicate that a slot in an object has been updated
   * @param oldTarget an Object reference indicating the old referent of the slot
   * @param o an Object reference indicating the objetc being updated; here
   *        null means a static variable is being updated
   * @param newTarget an Object reference indicating the new referent of the slot
   */
  private static native void _pointerMovedOff (Object oldTarget, Object o, Object newTarget);

  /**
   * native method to indicate that a field in a scalar object has been updated;
   * could also be used for arrays by passing the negative of the index
   * @param o an Object reference indicating the objetc being updated; here
   *        null means a static variable is being updated
   * @param newTarget an Object reference indicating the new referent of the slot
   * @param fieldId an int giving the unique number used the identify the field
   */
  //TODO: add native method for field update
  //private static native void _pointerUpdated (Object o, Object newTarget, int fieldId);

  /**
   * method to indicate that a field in a scalar object has been updated
   * @param o an Object reference indicating the objetc being updated
   * @param newTarget an Object reference indicating the new referent of the field
   * @param fieldId an int giving the unique number used the identify the field
   */
  /*
  public static void pointerUpdated (Object o, Object newTarget, int fieldId) {
    if (engaged == 1) {
      // TODO: _pointerUpdated(o, newTarget, fieldId);
    }
  }
  */

  /**
   * method to indicate that a static field has been updated
   * @param newTarget an Object reference indicating the new referent of the field
   * @param fieldId an int giving the unique number used the identify the field
   */
  /*
  public static void pointerUpdated (Object newTarget, int fieldId) {
    if (engaged == 1) {
      // TODO: _pointerUpdated(null, newTarget, fieldId);
    }
  }
  */

  /**
   * Indicates that a pointer slot has been updated.
   * NOTE: This swaps around the order of the parameters when it calls
   * the native method.  This is so that PUTFIELD and PUTSTATIC can both
   * use the same native method, and so that we add the minimal number of
   * bytecodes to the method.
   * @param newTarget an Object reference indicating the new referent of the slot
   * @param origin an Object reference indicating the objetc being updated; here
   *        null means a static variable is being updated
   * @param oldTarget an Object reference indicating the old referent of the slot
   */
  public static void pointerMovedOff (Object newTarget, Object origin, Object oldTarget) {
    if (engaged == 1) {
      _pointerMovedOff(oldTarget, origin, newTarget);
    }
  }

  /**
   * Indicates that a field of a scalar object has been updated, <b>and</b> indicates
   * the specific field.
   * @param newTarget an Object reference indicating the new referent of the slot
   * @param origin an Object reference indicating the objetc being updated; here
   *        null means a static variable is being updated
   * @param oldTarget an Object reference indicating the old referent of the slot
   * @param fieldId an int giving the unique number used the identify the field
   */
  public static void pointerMovedOff (Object newTarget, Object origin, Object oldTarget,
                                      int fieldId) {
    //For now, ignore fieldId
    if (engaged == 1) {
      _pointerMovedOff(oldTarget, origin, newTarget);
    }
  }

  /**
   * Indicates that a static field has been updated
   * @param newTarget an Object reference indicating the new referent of the slot
   * @param oldTarget an Object reference indicating the old referent of the slot
   */
  public static void pointerMovedOff (Object newTarget, Object oldTarget) {
    if (engaged == 1) {
      _pointerMovedOff(oldTarget, null, newTarget);
    }
  }

  /**
   * Indicates that a static field has been updated, <b>and</b> indicates
   * the specific field.
   * @param newTarget an Object reference indicating the new referent of the slot
   * @param fieldId an int giving the unique number used the identify the field
   */
  public static void pointerMovedOff (Object newTarget, Object oldTarget, int fieldId) {
    // Stack is ...,v,v,oldVal
    //For now, ignore fieldId
    if (engaged == 1) {
      _pointerMovedOff(oldTarget, null, newTarget);
    }
  }

  /**
   * indicates update to field of an uninitialized object
   * @param newTarget the Object being stored into the field
   * @param allocSite an int giving the unique allocation site of the object, which
   *        allows ET to associate this update with the right object, eventually
   * @param fieldId an int giving the unique id of the field being updated
   */
  private static native void _uninitPutfield (Object newTarget, int allocSite, int fieldId);

  /**
   * indicates update to field of an uninitialized object
   * @param newTarget the Object being stored into the field
   * @param allocSite an int giving the unique allocation site of the object, which
   *        allows ET to associate this update with the right object, eventually
   * @param fieldId an int giving the unique id of the field being updated
   */
  public static void uninitPutfield (Object newTarget, int allocSite, int fieldId) {
    if (engaged == 1) {
      _uninitPutfield(newTarget, allocSite, fieldId);
    }
  }

  /**
   * indicates that the given thread has allocated the given object;
   * called from Object.<init>, i.e., as soon as it is legal, or
   * immediate after NEWARRAY/ANEWARRAY for arrays (will change in future)
   * @param thread an Object indicating the allocating thread
   * @param o the newly allocated Object
   */
  private static native void _newobj (Object thread, Object o);

  /**
   * indicates a newly allocated object;
   * called from Object.<init>, i.e., as soon as it is legal
   * @param o the newly allocated Object
   */
  public static void newobj (Object o) {
    if (engaged == 1) {
      _newobj(Thread.currentThread(), o);
    }
  }
	
  /**
   * indicates that the given thread has allocated the given object,
   * and that it is of a subclass of java.lang.Reerence;
   * called from Object.<init>, i.e., as soon as it is legal, or
   * immediate after NEWARRAY/ANEWARRAY for arrays (will change in future)
   * @param thread an Object indicating the allocating thread
   * @param o the newly allocated Object
   */
  private static native void _newrefobj (Object thread, Object o);

  /**
   * indicates a newly allocated object, of a subclass of java.lang.Reference;
   * called from Object.<init>, i.e., as soon as it is legal
   * @param o the newly allocated Object
   */
  public static void newrefobj (Object o) {
    if (engaged == 1) {
      _newrefobj(Thread.currentThread(), o);
    }
  }
	
  /**
   * indicates that the given thread has allocated the given array
   * called immediately after NEWARRAY/ANEWARRAY
   * @param thread an Object indicating the allocating thread
   * @param o the newly allocated array Object
   */
  private static native void _newarray (Object thread, Object o);

  /**
   * indicates a newly allocated array
   * @param o the newly allocated array Object
   */
  public static void newarray (Object o) {
    if (engaged == 1) {
      _newarray(Thread.currentThread(), o);
    }
  }
	
  /**
   * indicates that the given thread has allocated the given object;
   * called at the NEW bytecode, so cannot pass the object;
   * rather, we identify it by its unique allocation site
   * @param thread an Object indicating the allocating thread
   * @param classId an int giving the unique id of the class of the new object
   * @param allocSite an int giving the unique allocation site of the new object
   */
  private static native void _newobj (Object thread, int classId, int AllocSite);

  /**
   * indicates a newly allocated object;
   * called at the NEW bytecode, so cannot pass the object;
   * rather, we identify it by its unique allocation site
   * @param classId an iiint giving the unique number of the class of the new object
   * @param allocSite an int giving the unique allocation site of the new object
   */
  public static void newobj (int classId, int allocSite) {
    if (engaged == 1) {
      _newobj(Thread.currentThread(), classId, allocSite);
    }
  }

  /**
   * indicates that the given thread has allocated the given object
   * and that it is of a subclass of Reference;
   * called at the NEW bytecode, so cannot pass the object;
   * rather, we identify it by its unique allocation site
   * @param thread an Object indicating the allocating thread
   * @param classId an int giving the unique id of the class of the new object
   * @param allocSite an int giving the unique allocation site of the new object
   */
  private static native void _newrefobj (Object thread, int classId, int AllocSite);

  /**
   * indicates a newly allocated object of a subclass of Reference;
   * called at the NEW bytecode, so cannot pass the object;
   * rather, we identify it by its unique allocation site
   * @param allocSite an int giving the unique allocation site of the new object
   */
  public static void newrefobj (int classId, int allocSite) {
    if (engaged == 1) {
      _newrefobj(Thread.currentThread(), classId, allocSite);
    }
  }

  /**
   * called when a Reference object is initialized, to tell the run-time
   * the Reference object and its referent; it is called just after the
   * java/lang/ref/Reference.<init> constructor returns to the subclass
   * constructor, and thus not called if an exception is thrown (which
   * would mean the Reference object was not initialized)
   * @param refObj a Java Reference object that is now initialized
   * @param referent the Object that is the Reference's referent
   */
  public static void refInit (Reference refObj, Object referent) {
    if (engaged == 1) {
      // TODO: call native method, once it is implemented
    }
  }

  /**
   * indicates a call to initialized an object
   * called just before the INVOKESPECIAL
   * @param thread an Object indicating the allocating thread
   * @param allocSite an int identifying the unique allocation site of the new object;
   *        0 means "uninitialized this"
   */
  private static native void _initOf (Object thread, int allocSite);

  /**
   * indicates a call to initialized an object
   * called just before the INVOKESPECIAL
   * @param allocSite an int identifying the unique allocation site of the new object;
   *        0 means "uninitialized this"
   */
  public static void initOf (int allocSite) {
    if (engaged == 1) {
      _initOf(Thread.currentThread(), allocSite);
    }
  }

  /**
   * used to mark entry to a method
   * @param methodId an int giving our unique number for the method
   * @param receiver an Object giving the receiver for an instance method;
   *        use null for a static method
   */
  private static native void _methodEntry (int methodId, Object receiver);

  /**
   * used to mark entry to a static method
   * @param methodId an int giving our unique number for the method
   */
  private static native void _methodEntryStatic (int methodId);

  /**
   * used to mark entry to a static method
   * @param methodId an int giving our unique number for the method
   */
  public static void methodEntryStatic (int methodId) {
    if (engaged == 1) {
      _methodEntryStatic(methodId);
    }
  }
	
  /**
   * used to mark entry to an instance method
   * @param methodId an int giving our unique number for the method
   * @param receiver an Object giving the receiver for an instance method;
   *        use null for an uninitialized object
   */
  public static void methodEntry (int methodId, Object receiver) {
    if (engaged == 1) {
      _methodEntry(methodId, receiver);
    }
  }
	
  /**
   * used to mark exit from a method
   * @param methodId an int giving our unique number for the method
   * @param receiver an Object giving the receiver for an instance method;
   *        use null for a static method
   */
  private static native void _methodExit (int methodId, Object receiver);

  /**
   * used to mark exit from a static method
   * @param methodId an int giving our unique number for the method
   */
  public static void methodExitStatic (int methodId) {
    if (engaged == 1) {
      _methodExit(methodId, null);
    }		
  }
	
  /**
   * used to mark exit from an instance method
   * @param methodId an int giving our unique number for the method
   * @param receiver an Object giving the receiver for the method
   */
  public static void methodExit (int methodId, Object receiver) {
    if (engaged == 1) {
      _methodExit(methodId, receiver);
    }
  }
	
  /**
   * used to mark exceptional exit from a static method
   * @param exception an Object giving the exception being thrown
   * @param methodId an int giving our unique number for the method
   */
  private static native void _methodExceptionStatic (Object exception, int methodId);

  /**
   * used to mark exceptional exit from a static method
   * @param exception an Object giving the exception being thrown
   * @param methodId an int giving our unique number for the method
   */
  public static void methodExceptionStatic (Object exception, int methodId) {
    if (engaged == 1) {
      _methodExceptionStatic(exception, methodId);
    }
  }
	
  /**
   * used to mark exceptional exit from an instance method
   * @param methodId an int giving our unique number for the method
   * @param receiver an Object giving the receiver for an instance method;
   *        use null for a static method
   */
  private static native void _methodException (Object exception, int methodId, Object receiver);

  /**
   * used to mark exceptional exit from an instance method
   * @param exception an Object giving the exception being thrown
   * @param methodId an int giving our unique number for the method
   * @param receiver an Object giving the receiver for the method;
   *        this will be null for a constructor
   */
  public static void methodException (Object exception, int methodId, Object receiver) {
    if (engaged == 1) {
      _methodException(exception, methodId, receiver);
    }
  }
	
  /**
   * indicates update of a slot in an array of objects
   * @param array the array Object being updated
   * @param i the int index being updated
   * @param storee the Object being stored into the slot
   */
  private static native void _arrayStore(Object array, int i, Object storee);

  /**
   * indicates update of a slot in an array of objects
   * NOTE: order is different than for the native method, to make the
   * bytecode sequence shorter in the instrumented code
   * @param i the int index being updated
   * @param storee the Object being stored into the slot
   * @param array the array Object being updated
   */
  public static void arrayStore (int i, Object storee, Object array) {
    if (engaged == 1) {
      _arrayStore(array,i,storee);
    }
  }

  /**
   * called just after allocating a multidimensional array
   * @param array the array Object just allocated
   * @param dims an int giving the number of dimensions of the array,
   *        which is necessarily > 0
   */
  public static void multiNewArray (Object[] array, int dims) {
    if (engaged == 1) {

      newobj(array);
      if (dims == 1) {
        // This can only happen if some one allocated
        // a single dimensional array with MULTIANEWARRAY,
        // which is legal, but odd.
        // In this case, we are all done.
        return;
      }

      for (int i = 0; i < array.length; i++) {
        if (dims > 2) {
          multiNewArray((Object[])array[i], dims-1);	
          pointerMovedOff((Object[])array[i], array, null);
        } else {
          // If it is just 2D, go through time stamping all the sub arrays.
          // This avoids having do do funny things with casts and instanceof
          newobj(array[i]);
          pointerMovedOff(array[i], array, null);
        }
      }
    }

  }

  /**
   * used to timestamp up to 5 objects
   * @param num an int indicating how many objects to timestamp;
   *        must lie in the range 1 .. 5; arguments past the given count are ignored
   * @param obj0 an Object timestamped if num >= 1 (i.e., always)
   * @param obj1 an Object timestamped if num > 1
   * @param obj2 an Object timestamped if num > 2
   * @param obj3 an Object timestamped if num > 3
   * @param obj4 an Object timestamped if num > 4
   */
  public static native void _varEvent (int num, Object obj0, Object obj1, Object obj2,
                                       Object obj3, Object obj4);

  /**
   * called to store an element in an array of Objects
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot being updated
   * @param val an Object giving the new value to store in the slot
   */
  public static void AAStoreEvent (Object[] array, int idx, Object val)
      throws NullPointerException, ArrayIndexOutOfBoundsException, ArrayStoreException {
    if (engaged == 1) {
      _arrayStore(array, idx, val);
    }
    array[idx] = val;
  }

  /**
   * called to store an element in an array of characters
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot being updated
   * @param val a char giving the new value to store in the slot
   */
  public static void CAStoreEvent (char[] array, int idx, char val)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    array[idx] = val;
  }

  /**
   * called to store an element in an array of shorts
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot being updated
   * @param val an int giving the new value to store in the slot
   */
  public static void SAStoreEvent (short[] array, int idx, int val)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    array[idx] = (short)val;
  }

  /**
   * called to store an element in an array of ints
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot being updated
   * @param val an int giving the new value to store in the slot
   */
  public static void IAStoreEvent (int[] array, int idx, int val)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    array[idx] = val;
  }

  /**
   * called to store an element in an array of longs
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot being updated
   * @param val a long giving the new value to store in the slot
   */
  public static void LAStoreEvent (long[] array, int idx, long val)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    array[idx] = val;
  }

  /**
   * called to store an element in an array of floats
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot being updated
   * @param val a float giving the new value to store in the slot
   */
  public static void FAStoreEvent (float[] array, int idx, float val)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    array[idx] = val;
  }

  /**
   * called to store an element in an array of doubles
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot being updated
   * @param val a double giving the new value to store in the slot
   */
  public static void DAStoreEvent (double[] array, int idx, double val)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    array[idx] = val;
  }

  /**
   * called to load an element from an array of characters
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot to load
   * @result the character in the slot
   */
  public static char CALoadEvent (char[] array, int idx)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    return array[idx];
  }

  /**
   * called to load an element from an array of shorts
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot to load
   * @result the short in the slot
   */
  public static short SALoadEvent (short[] array, int idx)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    return array[idx];
  }

  /**
   * called to load an element from an array of ints
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot to load
   * @result the int in the slot
   */
  public static int IALoadEvent (int[] array, int idx)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    return array[idx];
  }

  /**
   * called to load an element from an array of longs
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot to load
   * @result the long in the slot
   */
  public static long LALoadEvent (long[] array, int idx)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    return array[idx];
  }

  /**
   * called to load an element from an array of floats
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot to load
   * @result the float in the slot
   */
  public static float FALoadEvent (float[] array, int idx)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    return array[idx];
  }

  /**
   * called to load an element from an array of doubles
   * @param array an Object giving the array
   * @param idx an int giving the index of the slot to load
   * @result the double in the slot
   */
  public static double DALoadEvent (double[] array, int idx)
    throws NullPointerException, ArrayIndexOutOfBoundsException {
    if (engaged == 1) {
      _varEvent(1, array, null, null, null, null);
    }
    return array[idx];
  }

  /**
   * Timestamp one object
   * @param obj the Object to timestamp
   */
  public static void varEvent1 (Object obj) {
    if (engaged == 1) {
      _varEvent(1, obj, null, null, null, null);
    }
  }

  /**
   * Timestamp two objects
   * @param obj0 an Object to timestamp
   * @param obj1 an Object to timestamp
   */
  public static void varEvent2 (Object obj0, Object obj1) {
    if (engaged == 1) {
      _varEvent(2, obj0, obj1, null, null, null);
    }	
  }

  /**
   * Timestamp three objects
   * @param obj0 an Object to timestamp
   * @param obj1 an Object to timestamp
   * @param obj2 an Object to timestamp
   */
  public static void varEvent3 (Object obj0, Object obj1, Object obj2) {
    if (engaged == 1) {
      _varEvent(3, obj0, obj1, obj2, null, null);
    }
  }

  /**
   * Timestamp four objects
   * @param obj0 an Object to timestamp
   * @param obj1 an Object to timestamp
   * @param obj2 an Object to timestamp
   * @param obj3 an Object to timestamp
   */
  public static void varEvent4 (Object obj0, Object obj1, Object obj2, Object obj3) {
    if (engaged == 1) {
      _varEvent(4, obj0, obj1, obj2, obj3, null);
    }
  }

  /**
   * Timestamp four objects
   * @param obj0 an Object to timestamp
   * @param obj1 an Object to timestamp
   * @param obj2 an Object to timestamp
   * @param obj3 an Object to timestamp
   * @param obj4 an Object to timestamp
   */
  public static void varEvent5 (Object obj0, Object obj1, Object obj2, Object obj3, Object obj4) {
    if (engaged == 1) {
      _varEvent(5, obj0, obj1, obj2, obj3, obj4);
    }
  }

  /**
   * a main that does nothing interesting: just prints that it was called
   * @param argv the array of argument Strings passed to the program
   */
  public static void main (String[] argv) {
    System.out.println("Main called");
  }

}

// Local Variables:
// mode:Java
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
