#include "ExternalClassInstrumenter.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <iostream>
#include <fstream>
#include <errno.h>
#include <agent_util.h>
#include <java_crw_demo.h>
using namespace std;

ExternalClassInstrumenter::ExternalClassInstrumenter (jvmtiEnv *jvmti, string javapath,
                                                      string classpath, string jarpath,
                                                      string useOpt, string outputFile,
                                                      bool verbose, string experimental) {
  this->javapath = javapath;
  this->classpath = classpath;
  this->jarpath = jarpath;
  this->useOpt = useOpt;
  this->experimental = experimental;

  int pipeErr;

  // -- Set up the ASM subprocess

  pipeErr = pipe(parentToChild);
  if (pipeErr) {
    cerr << "ExternalClassInstrumenter::ExternalClassInstrumenter: could not pipe parentToChild."
         << endl;
    exit(1);
  }

  pipeErr = pipe(childToParent);
  if (pipeErr) {
    cerr << "ExternalClassInstrumenter::ExternalClassInstrumenter: could not pipe childToParent."
         << endl;
    exit(1);
  }

  childPid = fork();

  if (childPid < 0) {
    cerr << "ExternalClassInstrumenter::ExternalClassInstrumenter: Fork error." << endl;
    exit(1);
  }

  if (childPid) {
    // -- Parent
    //    Close unused ends of the pipe -- return to agent
    close(parentToChild[0]);
    close(childToParent[1]);

    // -- Create a lock

    jvmti->CreateRawMonitor("instrumenterLock", &lock);

  } else {
    // -- Child: start the ASM subprocess
    close(0);

    int dupErr = dup(parentToChild[0]);
    if (dupErr == -1) {
      cerr << "ExternalClassInstrumenter::ExternalClassInstrumenter: Could not dup." << endl;
      exit(1);
    }

    close(parentToChild[1]);

    close(1);
    dup(childToParent[1]);
    close(childToParent[0]);
    
    int execReturn;
    if (!verbose) {
      execReturn = execl(this->javapath.c_str(), this->javapath.c_str(),
                         "-classpath", this->classpath.c_str(), "-jar", this->jarpath.c_str(),
                         this->useOpt.c_str(), "-n", outputFile.c_str(),
                         this->experimental.c_str(), NULL);
    } else {
      execReturn = execl(this->javapath.c_str(), this->javapath.c_str(),
                         "-classpath", this->classpath.c_str(), "-jar", this->jarpath.c_str(),
                         this->useOpt.c_str(), "-n", outputFile.c_str(), "-v",
                         this->experimental.c_str(), NULL);
    }

    if (execReturn == -1) {
      //Do I even need to check the condition? If we got here something screwed up
      cerr << "Exec failed." << endl;
      cerr << "Command line was: " << javapath << "  -classpath " << classpath
           << " -jar " << jarpath << "-n" << outputFile.c_str() << experimental << endl;
      exit(-1);
    }
  }
}

ExternalClassInstrumenter::~ExternalClassInstrumenter () {

  int closeErr;

  int status;
  closeErr = close(parentToChild[1]);
  if (closeErr) {
    cerr << "ExternalClassInstrumenter::~ExternalClassInstrumenter: "
         << "Could not close parent to child pipe." << endl;
    exit(1);
  }

  closeErr =close(childToParent[0]);
  if (closeErr) {
    cerr << "ExternalClassInstrumenter::~ExternalClassInstrumenter: "
         << "Could not close childToParent pipe." << endl;
    exit(1);
  }
	
  wait(&status);
}

void ExternalClassInstrumenter::instrumentClass(jvmtiEnv *jvmti,
						JNIEnv* env,
						jclass class_being_redefined,
						jobject loader,
						const char* name,
						jobject protection_domain,
						jint class_data_len,
						const unsigned char* class_data,
						jint* new_class_data_len,
						unsigned char** new_class_data) {
  string command;
  unsigned char* buffer;
  size_t bytesRead;
  size_t bytesWritten;
  int childExitStatus;
  int execReturn;
  pid_t childPid;
  jvmtiError error;

  error = jvmti->RawMonitorEnter(lock);
  check_jvmti_error(jvmti, error, "ExternalClassInstrumenter::lockOn: Cannot enter with raw monitor");

  // buffer = (unsigned char*)malloc(sizeof(char)*class_data_len*10); //We are hoping we don't need more than this.

  // cout << "(1c) Write out " << (int)class_data_len << " bytes for class " << name << endl;
  writeUntilDone(parentToChild[1], class_data, (int)class_data_len);
  // close(parentToChild[1]);//Pipe is fully closed, so child will read EOF

  //bytesRead = read(childToParent[0], buffer, (size_t)(class_data_len*2));
  // cout << "(3c) Read in transformed class " << name << endl;
  int new_length = 0;
  unsigned char * new_bytes = readUntilDone(jvmti, childToParent[0], new_length);
  // close(childToParent[0]);
  // cout << "(8c) ...got " << new_length << " bytes" << endl;

  if (new_length == 0) {
    cerr << "Could not read from pipe." << endl;
    exit(1);
  }

  // waitpid(childPid, &childExitStatus, 0);
  // *new_class_data_len = (jint)bytesRead;
  // *new_class_data = (unsigned char*)malloc(bytesRead);
  // memcpy(*new_class_data, buffer, bytesRead);
  // free(buffer);

  (*new_class_data) = new_bytes;
  (*new_class_data_len) = new_length;

  // -- Check
  //bool ok = true;
  //for (int i = 0; i < new_length; i++) {
  //  if (class_data[i] != new_bytes[i])
  //    ok = false;
  //}

  //if (class_data_len != new_length)
  //  ok = false;

  // cout << "(10c) Result: " << (ok ? "GOOD" : "BAD") << endl;

  error = jvmti->RawMonitorExit(lock);
  check_jvmti_error(jvmti, error, "ExternalClassInstrumenter::UnlockOn:Cannot exit with raw monitor");
}

void ExternalClassInstrumenter::writeUntilDone (int fd, const unsigned  char* buffer, int length) {
  int byteIndex = 0;
  int totalBytesWritten = 0;

  // cout << "(2c) writeUntilDone: length = " << length << endl;

  /*
  cout << "C Write Data:" << endl;
  for (int i = 0; i < length; i++) {
    cout << "  " << i << " : " << (unsigned int)buffer[i] << endl;
  }
  */

  writeSize(length, fd);
  writeFully(fd, buffer, length);
}

unsigned char *ExternalClassInstrumenter::readUntilDone (jvmtiEnv *jvmti, int fd, int & length) {
  int bytesLastRead = 0;
  int totalBytesRead = 0;
  int bytes_to_read = 0;
  jvmtiError error;

  // -- Read number of bytes

  bytes_to_read = readSize(fd);
  // cout << "(6c) readUntilDone: bytes_to_read = " << bytes_to_read << endl;

  // -- Allocate buffer
  unsigned char* buffer;
  //Cannot use malloc,
  //We need to use jvmti->Alloc, because the VM may try to free this memory.
  error = jvmti->Allocate(bytes_to_read,&buffer);
  check_jvmti_error(jvmti, error,
    "ExternalClassInstrumenter:: readUntilDone: could not allocate memory for instrumented class.");

  readFully(fd, buffer, bytes_to_read);
  /*
  cout << "C Read Data:" << endl;
  for (int i = 0; i < bytes_to_read; i++) {
    cout << "  " << i << " : " << (unsigned int)buffer[i] << endl;
  }
  */

  length = bytes_to_read;
  return buffer;
}

unsigned int ExternalClassInstrumenter::readSize (int fd) {
  // -- First write number of bytes 
  //    (NOTE: must be compatible with DataOutputStream)
  unsigned char sizeBytes[4];
  readFully(fd, sizeBytes, 4);

  unsigned int bytes_to_read = 
       ((int)sizeBytes[3]) 
    + (((int)sizeBytes[2]) << 8)
    + (((int)sizeBytes[1]) << 16)
    + (((int)sizeBytes[0]) << 24);

  return bytes_to_read;
}

void ExternalClassInstrumenter::writeSize (unsigned int size, int fd) {
  // -- First write number of bytes 
  //    (NOTE: must be compatible with DataInputStream)
  unsigned char sizeBytes[4];
  sizeBytes[0] = (size & 0xFF000000) >> 24;
  sizeBytes[1] = (size & 0x00FF0000) >> 16;
  sizeBytes[2] = (size & 0x0000FF00) >> 8;
  sizeBytes[3] =  size & 0x000000FF;
  writeFully(fd, sizeBytes, 4);
}

void ExternalClassInstrumenter::readFully (int fd, unsigned char * buffer, int length) {
  // -- Read until done
  int bytes_left = length;
  int bytes_read = 0;
  int total_read = 0;
  while (bytes_left > 0) {
    bytes_read = read(fd, (buffer + total_read), bytes_left);
    bytes_left = bytes_left - bytes_read;
    total_read = total_read + bytes_read;
  }
}

void ExternalClassInstrumenter::writeFully (int fd, const unsigned char * buffer, int length) {
  // -- Write until done
  int bytes_left = length;
  int bytes_read = 0;
  int total_read = 0;
  while (bytes_left > 0) {
    bytes_read = write(fd, (buffer + total_read), bytes_left);
    bytes_left = bytes_left - bytes_read;
    total_read = total_read + bytes_read;
  }
}

// Local Variables:
// mode:C++
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
