#ifndef FLATTRACEOUTPUTTER_H_
#define FLATTRACEOUTPUTTER_H_

#include "CallBackHandler.h"
#include "TraceOutputter.h"
#include <string>
#include "jni.h"
#include "Record.h"
#include <map>
#include <vector>
using namespace std;

#define DEFAULT_MAX_RECORDS 20000000

/*struct record {
	string output;
	long time;	
};*/
	

class FlatTraceOutputter : public TraceOutputter
{
public:
	FlatTraceOutputter(jvmtiEnv* jvmti, ostream *out, void (*fullcb)(void));
	FlatTraceOutputter(jvmtiEnv* jvmti, ostream *out, unsigned int recordLimit,  void (*fullcb)(void));
	virtual ~FlatTraceOutputter();
	virtual void doMethodEntry(unsigned long time, int methodId, /* string className, string methodName, string signature, */ unsigned long receiver, unsigned long thread);
	virtual void doMethodExit(unsigned long time, int methodId, /* string className, string methodName, string signature, */  unsigned long receiver, unsigned long thread);
	virtual void doAllocation(unsigned long time, int objectTag, jlong size, string type, unsigned long thread);
	
	#ifdef THREAD_TAG_UPDATES
	virtual void doPointerUpdate(unsigned long time, long oldTarget, long origin, long newTarget, long thread);
	#else
	virtual void doPointerUpdate(unsigned long time, long oldTarget, long origin, long newTarget);
	#endif
	virtual void doVarEvent(unsigned long time, unsigned long obj, unsigned long thread);
	virtual void doDeath(long tag, long deathTime);
	virtual void doGCEnd();
	virtual void doRoot(unsigned long time,  unsigned long obj);
	//virtual void flush();
	
private:
    ostream* output;
    // map<string,bool> methodSet;
    // vector<Record*> outputBuffer;
    void initialize(void (*fullcb)(void), unsigned int recordLimit);

    Record ** outputBuffer;
    Record ** deathBuffer;
    //We keep some slack at the end of the output buffer
    //because we copy death records there to sort.
    //Real buffer size represents the actual allocated size.
    unsigned int realBufferSize;
    unsigned int deathBufferSize;
    //Pointer into outputBuffer, where we can add a new record
    volatile unsigned int bufferEnd;
    volatile unsigned int deathBufferEnd;
    //How many records have we finished adding to the buffer?
    volatile unsigned int recFinishedCount;
    unsigned int maxRecords; //Do a call back if we exceed this number of records
    void (*fullcb)(void);


    vector<char> debugList;
    // long currentTime;
    jrawMonitorID  yieldLock;
    jrawMonitorID  fullMon;
    jrawMonitorID  stdoutLock;
    void addRecordToList(Record* rec);
    void addRecordToDeathList(Record* rec);
    jvmtiEnv*  aJVMTI;


    volatile unsigned int gcsStarted;
    volatile unsigned int gcsFinished;
  

    // -- Counters
    unsigned long countMethodEntries;
    unsigned long countMethodExits;
    unsigned long countAllocations;
    unsigned long countPointerUpdates;
    unsigned long countDeaths;
    unsigned long countVarEvents;
};

#endif /*FLATTRACEOUTPUTTER_H_*/
