/*
 * HashGraph.h
 *
 *  Created on: Dec 17, 2009
 *      Author: nricci01
 */

#ifndef HASHGRAPH_H_
#define HASHGRAPH_H_

#include <set>
#include <map>
#include <vector>
#include <utility>
#include <string>

using namespace std;


/*This is a graph data structure for use by ElephantTracks.
  It is not a fully general graph structure, because it makes some assumptions about how
  it will be used.
  In particular, removeNode(n) does not clean up any incoming edges to n, just the out going edges.
  We can get away with this in ET because only dead objects are ever removed, and if n is dead, anything 
  pointing to n also must be dead, and therefore the incoming edge will eventually be removed.
  This saves time/space tracking incoming edges. */

class HashGraph {
public:
	HashGraph();
	virtual ~HashGraph();
	void addEdge(long,long);
	bool removeEdge(long,long);
	bool removeNode(long);
	pair< vector<long>::iterator, vector<long>::iterator > neighbors(long target);
	string toString() const;
	vector<long> getNodes() const;
private:
	map<unsigned long, vector<long> > edgeMap;
	vector<long>::iterator findEdge(vector<long>& edgeList, long target);

};

#endif /* HASHGRAPH_H_ */
