package classReWriter;

import java.io.*;
import org.objectweb.asm.*;
import org.objectweb.asm.util.*;
import java.util.HashMap;
import java.util.Map;

public class ClassReWriter {

  private static Map<String,String> argMap;
  private static int methodIdCounter = 0;
  private static PrintStream nameStream = null; //For outputting the method ID->Name mapping.
  private static boolean verbose = false;
	
  public static int getNextMethodId() {
    methodIdCounter++;
    return methodIdCounter;
  }

  public static void verbosePrintln (String s) {
    if (verbose) {
      System.err.println(s);
    }
  }

  public static void main (String[] argv) throws IOException {
    argMap = parseCommandLineArgs(argv);

    if (argMap.containsKey("-n")) {
      File nameFile = new File(argMap.get("-n"));
      nameStream = new PrintStream(new FileOutputStream(nameFile) );
    } else {
      nameStream = new PrintStream(System.out);
    }

    if (argMap.containsKey("-v")) {
      verbose = true;
    }

    while (instrumentOneClass()) {}
  }
	
  private static Map<String,String> parseCommandLineArgs (String[] args) {
    Map<String,String> argMap = new HashMap<String,String>();
		
    for (int i =0; i < args.length; i++) {
      if (args[i].equals( "-c")) {
	argMap.put("-c", args[++i]);
      } else if (args[i].equals("-s")) {
	argMap.put("-s", null);
      } else if (args[i].equals ("-n")) {
	argMap.put("-n", args[++i]);
      } else if (args[i].equals ("-v")) {
	argMap.put("-v", null);
      } else if (args[i].equals("-opt")) {
        argMap.put("-opt", null);
      } else if (args[i].startsWith("-x")) {
        argMap.put("-x", args[i].substring(2));
      } else if (args[i].equals("-i")) {
        argMap.put("-i", args[++i]);
      } else {
	//TODO: Throw some kinda exception...
      }		
    }
    return argMap;
  }
	
  private static boolean instrumentOneClass () {
    byte [] originalClassBytes;

    try {
      // -- Read in the class
      int size_in;
      DataInputStream dataIn;

      if (!argMap.containsKey("-c")) {
        dataIn = new DataInputStream(System.in);

        // -- Read in the size
        //    NOTE: must be compatible with ExternalClassInstrumenter::writeUntilDone
        try {
          size_in = dataIn.readInt();
        } catch (IOException e) {
          verbosePrintln("ClassReWriter: input stream closed");
          return false;
        }
      } else {
	//We will read in the class specified by the -c parameter
	File c  = new File(argMap.get("-c"));
	size_in = (int)c.length();
	dataIn = new DataInputStream(new FileInputStream(c));
     }

      // -- Read in the class bytes
      originalClassBytes = new byte[size_in];
      dataIn.readFully(originalClassBytes);
    } catch (java.io.IOException e) {
      //System.err.println(e.getStackTrace().toString());

      verbosePrintln("ClassReWriter: could not read in class bytes: " + e.getMessage());
      return false;
    }

    // System.err.println("(4j) ClassReWriter: read in " + originalClassBytes.length + " bytes");

    /*
      int count = 0;
      System.err.println("Java Read Data:");
      for (byte b:originalClassBytes ){
      System.err.println("  " + (int)count + " : " + (int)b);
      count++;
      }
    */

    // -- Transform the class

    boolean optimizing = argMap.containsKey("-opt");
    boolean experimental = argMap.containsKey("-x");
    String expFlags = (experimental ? argMap.get("-x") : null);
    if (optimizing) {
      // pre-process to get frames
      ClassReader cr = new ClassReader(originalClassBytes);
      ClassWriter cw = new ETClassWriter(ClassWriter.COMPUTE_FRAMES);
      ETPreprocessor etp = new ETPreprocessor(cw);
      cr.accept(etp, ClassReader.EXPAND_FRAMES);
      originalClassBytes = cw.toByteArray();
      if (argMap.containsKey("-i")) {
        String intName = argMap.get("-i");
        try {
	  DataOutputStream dout = new DataOutputStream(new FileOutputStream(intName));
	  dout.write(originalClassBytes, 0, originalClassBytes.length);
	  dout.close();
        } catch (IOException exc) {
          System.err.printf("Problem writing intermediate file %s%n", intName);
        }
      }
    }
    ClassReader cr = new ClassReader(originalClassBytes);
    ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
    ETAdapter mearthAdapter = new ETAdapter(cw, optimizing, expFlags, nameStream);

    byte [] instrumentedClassBytes;    
      
    try{
      int crOptions = (optimizing ? (ClassReader.EXPAND_FRAMES | ClassReader.SKIP_DEBUG)
                                  : (ClassReader.EXPAND_FRAMES | ClassReader.SKIP_FRAMES));
      /*
      OutputStream os = new FileOutputStream("ET-trace.txt");
      PrintWriter pw = new PrintWriter(os, true);
      TraceClassVisitor tcv = new TraceClassVisitor(mearthAdapter, pw);
      */
      cr.accept(mearthAdapter,crOptions);
    }
    /*
    catch(java.lang.ArrayIndexOutOfBoundsException e){
      System.err.println(e.getStackTrace().toString());
      System.err.println("javaClassReWriter: main(): Caught Array index out of bounds exception " + e.getMessage());
      System.exit(2);
    }
    */
    /*
    catch(FileNotFoundException e){
	System.err.println("Could not open trace file:\n" + e);
    }
    */
    catch (java.lang.IllegalStateException e) {
      try {
        //e.printStackTrace();
        System.err.println("javaClassReWriter: InstrumentOneClass(): Caught IllegalStateException " +
                           e.getMessage());
        System.err.println("Class bytes written to IllegalStateDump.class");
        FileOutputStream dumpFile = new FileOutputStream("IllegalStateDump.class");

        dumpFile.write(originalClassBytes);
        dumpFile.close();
      } catch(Exception e2) {
      }
      System.exit(2);
    }

    instrumentedClassBytes = cw.toByteArray();

    int size_out = instrumentedClassBytes.length;

    // System.err.println("(5j) ClassReWriter: write out " + instrumentedClassBytes.length + " bytes");

    /*
    System.err.println("Java Write Data:");
    for (byte b:instrumentedClassBytes ){
      System.err.println("  " + (int)count + " : " + (int)b);
      count++;
    }
    */
      
    try {
      // -- Write the resulting class

      DataOutputStream dataOut = new DataOutputStream(System.out);
      if (!argMap.containsKey("-c")) {
        dataOut.writeInt(size_out);
      }
      dataOut.write(instrumentedClassBytes, 0, size_out);
      dataOut.flush();
    } catch (java.io.IOException e) {
      System.err.println(e.getStackTrace().toString());
      System.err.println("ClassReWriter: could not write out class bytes: " + e.getMessage());
      return false;
    }

    // System.err.println("(9j) ClassReWriter: Done");
    if (argMap.containsKey("-c")) {
      return false;
    } else {
      return true;
    }
  }
}

// Local Variables:
// mode:Java
// c-basic-offset:2
// indent-tabs-mode:nil
// End:
