package classReWriter;

import org.objectweb.asm.*;
import org.objectweb.asm.commons.*;
import java.util.Vector;
import java.io.*;

public class ETMethodAdapter extends AdviceAdapter {
  //Are we inside a method in a place that is safe to instrument
  //In particular, this is used to avoid inserting code in a
  //constructor before the call to the super constructor
  //which apparently causes the gates of Hell to open
  //spewing forth demons into the JVM
	
  private boolean insideMethod = false;
  private boolean foundEntry = false;
  private static final boolean STACK_SCAN_ON_ENTRY = true; 


	
  //private String awaitingConstructor;
		
  /*Name of the method this ETMethodAdapter is working on */
  protected String methodName;
  protected String methodDesc;
  protected String className;
  protected int  access;

  protected int methodId;
  
	
  public ETMethodAdapter(int access, String className, String name, String desc, MethodVisitor mv, PrintStream nameStream)
  {
    super(mv, access, name, desc);
    this.className = className;
    this.methodName = name;
    this.methodDesc = desc;
    this.access =  access;

    methodId = ClassReWriter.getNextMethodId();

    nameStream.println("N 0x" + Integer.toHexString(methodId) + " " + className + " " + methodName + " " + methodDesc);
  }

  /*Look through the signature to determine which local var slots are the parameters and object references  */
  public Vector<Integer> findParamSlots(String desc){
	Vector<Integer> slots = new Vector<Integer>();
	int slotIndex = 0;//offset for double wide parameters

	if( (access & Opcodes.ACC_STATIC) == 0 ){
	      //deal with the self parameter for non-static methods
              slots.add(0); // timestamp here so show live after method entry
	      slotIndex+= 1;
	}


	//start i at 1 to skip the left-paren '('
	for(int i = 1; i < desc.length(); i++ ){

		if(desc.charAt(i) == ')'){
			break;
		}

	

		if(desc.charAt(i) == 'D' || desc.charAt(i) == 'J'){
			//Doubles, longs take up an extra slot
			slotIndex++;	
		}

		if(desc.charAt(i) == 'L'){
			slots.add(slotIndex);
			//advance i past the end of this part of the descriptor
			// Looks like Ljava/lang/Object;
			while(desc.charAt(i) != ';'){
				i++;
			}
		}

		if(desc.charAt(i) == '['){
			slots.add(slotIndex);
			//If the descriptor is something like [[[[char
                        //we only want to return a slot index for the outer array
			while(desc.charAt(i) == '['){
				i++;
			}

			//It might also be something like  [[[[Ljava/lang/Object;, then we need to advance to the semi colon
			if(desc.charAt(i) == 'L'){
				while(desc.charAt(i) != ';'){
					i++;
				}
			}
				

		}
		
		slotIndex++;
	}

	return slots;
  }

  //Based on the given descriptor, does this method return an object/array?
  public boolean returnsObject(String desc){
	int cparenI; //Index in desc of the closing paren character ')'

	for(cparenI = 0; desc.charAt(cparenI) != ')'; cparenI++){
	}

	if(desc.charAt(cparenI + 1) == 'L' || desc.charAt(cparenI +1) == '['){
		return true;
	}
	else{
		return false;
	}
  }


    @Override
    public void visitCode(){
        // // Draft of code for putting a handler around each method
        // Label startLabel   = new Label();
        // Label endLabel     = new Label();
        // Label handlerLabel = new Label();
        // mv.visitTryCatchBlock(startLabel, endLabel, handlerLabel, null);
        // mv.visitLabel(startLabel);
	boolean inObject = className.equals("java/lang/Object");
	if (inObject && methodName.equals("finalize")){
	    // don't instrument finalize() -- causes infinite recursion!
	    mv.visitCode();
	    return;
	}
	if (inObject && methodName.equals("<init>")){
	    // Note new object
            mv.visitVarInsn(Opcodes.ALOAD, 0);
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "newobj", "(Ljava/lang/Object;)V");
        }
        super.visitCode();
        // mv.visitLabel(endLabel);
        // mv.visitLabel(handlerLabel);
        // mv.visitLdcInsn(methodId);
	// if( (access & Opcodes.ACC_STATIC) != 0){
        //     mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExitStatic", "(I)V");
        // }else{
        //     mv.visitVarInsn(Opcodes.ALOAD, 0);
        //     mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExit",
        //                        "(ILjava/lang/Object;)V");
        // }
        // mv.visitInsn(Opcodes.ATHROW);
    }


   @Override
    public void visitFieldInsn(int opcode, String owner, String fieldName, String fieldDesc){
    if (insideMethod){
      boolean isObj = isObjectReference(fieldDesc);
      switch(opcode){
      case Opcodes.PUTFIELD:
	  //Stack must look like ...,o,v since we are just before a put field
	  if( isObj ){
	      // want to call pointerMovedOff
	      mv.visitInsn(Opcodes.DUP2); // ...,o,v,o,v;
	      mv.visitInsn(Opcodes.SWAP); // ...,o,v,v,o
	      mv.visitInsn(Opcodes.DUP);  // ...,o,v,v,o,o
	      mv.visitFieldInsn(Opcodes.GETFIELD, owner, fieldName, fieldDesc); // ...,o,v,v,o,old
	      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "pointerMovedOff", "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V");
	  }else{
	      //want to call varEvent1 on the object
	      char kind = fieldDesc.charAt(0);
	      if( kind == 'J' || kind == 'D' ){
		  // there are two slots on top of the stack
		  mv.visitInsn(Opcodes.DUP2_X1);  // ...,wide,o,wide
		  mv.visitInsn(Opcodes.POP2);     // ...,wide,o
		  mv.visitInsn(Opcodes.DUP_X2);   // ...,o,wide,o
		  mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
	      }else{
		  // there is one slots on top of the stack
		  mv.visitInsn(Opcodes.DUP_X1);   // ...,v,o,v
		  mv.visitInsn(Opcodes.POP);      // ...,v,o
		  mv.visitInsn(Opcodes.DUP_X1);   // ...,o,v,o
		  mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
	      }
	  }
	  break;
		  
      case Opcodes.PUTSTATIC:
	  if( isObjectReference(fieldDesc) ){
	      // Stack is ...,v;
	      mv.visitInsn(Opcodes.DUP);
	      // Stack is ...,v,v;
	      mv.visitFieldInsn(Opcodes.GETSTATIC, owner, fieldName, fieldDesc);
	      // Stack is ...,v,v,oldVal
	      //mv.visitInsn(Opcodes.SWAP);
	      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "pointerMovedOff", "(Ljava/lang/Object;Ljava/lang/Object;)V");
	      // Stack is ...,v;
	  }
	  break;

      case Opcodes.GETFIELD:
	  mv.visitInsn(Opcodes.DUP);
	  mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
	  break;
      }	
    }
	
    super.visitFieldInsn(opcode, owner, fieldName, fieldDesc);
  }
		
  @Override
    public void visitInsn(int opcode){
    // System.err.println("visitInsn: " + className + "::" + methodName);
    if(insideMethod){
      /*
	if( (opcode >= Opcodes.IRETURN && opcode <= Opcodes.RETURN) || opcode == Opcodes.ATHROW ){
	System.err.println("instrumentExit: " + className + "::" + methodName);
	mv.visitLdcInsn(className);
	mv.visitLdcInsn( methodName);
	mv.visitLdcInsn( methodDesc);
			
	if( (access & Opcodes.ACC_STATIC) != 0){
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExitStatic", "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V");
	}
	else{
	mv.visitVarInsn(Opcodes.ALOAD, 0);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExit", "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V");
	}
	}
      */
			
      if(opcode == AASTORE){
        if(true){ // new version
          mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "AAStoreEvent", "([Ljava/lang/Object;ILjava/lang/Object;)V");
          return;
        }
	// Stack is a,idx,val);
	mv.visitInsn(Opcodes.DUP2_X1);
	// Stack is idx,val,a,idx,val
	mv.visitInsn(Opcodes.DUP2_X1);
	// Stack is idx,val,idx,val,a,idx,val
	mv.visitInsn(Opcodes.POP2);
	// Stack is idx,val,idx,val,a
	mv.visitInsn(Opcodes.DUP_X2);
	// Stack is idx,val,a,idx,val,a
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "arrayStore", "(ILjava/lang/Object;Ljava/lang/Object;)V");
	// Stack is idx,val,a
	mv.visitInsn(Opcodes.DUP_X2);
	// Stack is a,idx,val,a
	mv.visitInsn(Opcodes.POP);
	// Stack is a,idx,val
					
      }else if(opcode == Opcodes.AALOAD || opcode == Opcodes.BALOAD || opcode == Opcodes.CALOAD ||
               opcode == Opcodes.DALOAD || opcode == Opcodes.FALOAD || opcode == Opcodes.IALOAD ||
               opcode == Opcodes.LALOAD || opcode == Opcodes.SALOAD){
	 // Stack is a,idx
         if(opcode == Opcodes.CALOAD){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "CALoadEvent", "([CI)C");
            return;
	 }else if(opcode == Opcodes.FALOAD){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "FALoadEvent", "([FI)F");
            return;
	 }else if(opcode == Opcodes.IALOAD){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "IALoadEvent", "([II)I");
            return;
	 }else if(opcode == Opcodes.SALOAD){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "SALoadEvent", "([SI)S");
            return;
	 }else if(opcode == Opcodes.LALOAD){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "LALoadEvent", "([JI)J");
            return;
	 }else if(opcode == Opcodes.DALOAD){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "DALoadEvent", "([DI)D");
            return;
         }else{ // use for BALOAD since combines byte and boolean, for AALOAD because of return type
            mv.visitInsn(Opcodes.DUP2);
            // Stack is a,idx,a,idx
            mv.visitInsn(Opcodes.POP);
            // Stack is a,idx,a
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
         }
	 // Stack is a,idx


      }else if(opcode == Opcodes.BASTORE || opcode == Opcodes.CASTORE ||
               opcode == Opcodes.FASTORE || opcode == Opcodes.IASTORE || opcode == Opcodes.SASTORE){
         // arrayref, idx, val
         if(opcode == Opcodes.CASTORE){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "CAStoreEvent", "([CIC)V");
            return;
	 }else if(opcode == Opcodes.FASTORE){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "FAStoreEvent", "([FIF)V");
            return;
	 }else if(opcode == Opcodes.IASTORE){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "IAStoreEvent", "([III)V");
            return;
	 }else if(opcode == Opcodes.SASTORE){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "SAStoreEvent", "([SII)V");
            return;
	 }else{ // must use this for BASTORE since includes boolean and byte
            mv.visitInsn(Opcodes.DUP2_X1);  // idx, val, arrayref, idx, val
            mv.visitInsn(Opcodes.POP2);     // idx, val, arrayref
            mv.visitInsn(Opcodes.DUP_X2);   // arrayref, idx, val, arrayref
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
	 }
         // arrayref, idx, val


      }else if(opcode == Opcodes.DASTORE || opcode == Opcodes.LASTORE){
         // arrayref, idx, wideval
         if(opcode == Opcodes.DASTORE){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "DAStoreEvent", "([DID)V");
            return;
         }else if(opcode == Opcodes.LASTORE){
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "LAStoreEvent", "([JIJ)V");
            return;
	 }else{ // this branch now dead
            mv.visitInsn(Opcodes.DUP2_X2); // wideval, arrayref, idx, wideval
            mv.visitInsn(Opcodes.POP2);    // wideval, arrayref, idx
            mv.visitInsn(Opcodes.DUP2_X2); // arrayref, idx, wideval, arrayref, idx
            mv.visitInsn(Opcodes.POP);     // arrayref, idx, wideval, arrayref
            mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
         }
         // arrayref, idx, wideval
      }else if(opcode == Opcodes.ARRAYLENGTH || opcode == Opcodes.ATHROW){
         mv.visitInsn(Opcodes.DUP);
         mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
      }
    }
    super.visitInsn(opcode);	
  }
	
  @Override
  public void visitJumpInsn(int opcode, Label l){
	if(opcode == Opcodes.IF_ACMPEQ || opcode == Opcodes.IF_ACMPNE){
		mv.visitInsn(Opcodes.DUP2);
		mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent2", "(Ljava/lang/Object;Ljava/lang/Object;)V");
	}
	if(opcode == Opcodes.IFNULL || opcode == Opcodes.IFNONNULL){
		mv.visitInsn(Opcodes.DUP);
		mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
	}
	mv.visitJumpInsn(opcode,l);
  }

  @Override
  public void visitIntInsn(int opcode, int operand){
    super.visitIntInsn(opcode, operand);
    if(opcode == Opcodes.NEWARRAY){
      mv.visitInsn(Opcodes.DUP);
      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "newobj", "(Ljava/lang/Object;)V");
    }
			
  }

  /*
  @Override
    public void visitTryCatchBlock(Label start, Label end, Label handler, String type)
  {
    if (type == null) {
      System.err.println("In finally block");
      inFinally = true;
    }
    super.visitTryCatchBlock(start, end, handler, type);
  }
  */


  @Override
  public void visitLdcInsn(Object obj){
    super.visitLdcInsn(obj);
    // if(obj instanceof String || obj instanceof Type){
	// mv.visitInsn(Opcodes.DUP);
	// mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
    // }
  }

  @Override
  public void visitMethodInsn(int opcode, String owner, String name, String desc){
	super.visitMethodInsn(opcode, owner,name,desc);

	if(returnsObject(desc)){//This indicates the method returns an object
		mv.visitInsn(Opcodes.DUP);
		mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
	}	
	

  }

  @Override
  public void visitMultiANewArrayInsn(String desc, int dims){
	super.visitMultiANewArrayInsn(desc, dims);
	mv.visitInsn(Opcodes.DUP);
	//This requires that dims actually be a short.
	//I hope no one wants an array with more than
	// 32,767 dimensions
	mv.visitIntInsn(Opcodes.SIPUSH, dims);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "multiNewArray", "([Ljava/lang/Object;I)V");
  }

  @Override
  public void visitTypeInsn(int opcode, String type){
	if(opcode == Opcodes.ANEWARRAY){
	      super.visitTypeInsn(opcode,type);
	      mv.visitInsn(Opcodes.DUP);
      	      mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "newobj", "(Ljava/lang/Object;)V");	
	}
	else if(opcode == Opcodes.INSTANCEOF || opcode == Opcodes.CHECKCAST){
	     mv.visitInsn(Opcodes.DUP);
      	     mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
	     super.visitTypeInsn(opcode,type);
	}
	else{
	     super.visitTypeInsn(opcode,type);
	}
  }

/*
  @Override
  public void visitVarInsn(int opcode, int var)
  {
    if (false // ! className.startsWith("java/lang/System")
	// ! className.startsWith("java")
	) {
      // className.startsWith("spec/jbb/infra")) {
      if (insideMethod && opcode == ASTORE && (var > 0)) {
	// System.err.println("GOT ONE");
	// mv.visitVarInsn(Opcodes.ALOAD, var);
	// mv.visitInsn(Opcodes.POP);
	// mv.visitInsn(ACONST_NULL);
	
	mv.visitVarInsn(Opcodes.ALOAD, var);
	// mv.visitInsn(Opcodes.DUP);
	mv.visitInsn(Opcodes.POP);
	// mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent", "(Ljava/lang/Object;)V");
      }
    }
    super.visitVarInsn(opcode, var);
  }
*/
		
  @Override 
    protected void onMethodEnter(){
    insideMethod = true;
    foundEntry = true;
    // System.err.println("onMethodEnter: " + className + "::" + methodName);
    // System.err.println("onMethodEnter: " + className + "::" + methodName);

    if(STACK_SCAN_ON_ENTRY){
      // -- Emit method ID
      mv.visitLdcInsn(methodId);
					
      if( (access & Opcodes.ACC_STATIC) != 0){
	// System.err.println("ADD-ENTRY: " + className + "::" + methodName);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodEntryStatic", 
			   "(I)V" );
			   /* "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"*/
      }
      else{
	// System.err.println("ADD-ENTRY: " + className + "::" + methodName);
	mv.visitVarInsn(Opcodes.ALOAD, 0);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodEntry", 
			   "(ILjava/lang/Object;)V");
			   /* "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"*/
      }
    }

    Vector<Integer> paramSlots = findParamSlots(methodDesc);

    /*
    for(int ps: paramSlots){
	mv.visitVarInsn(Opcodes.ALOAD, ps);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent", "(Ljava/lang/Object;)V");
    }*/

    //load all the parameters to the operand stack
    
    for(int ps: paramSlots){
	mv.visitVarInsn(Opcodes.ALOAD, ps);
    }
    
    //In case there are a lot of parameters, stamp them using as few calls to varEvent 5 as possible.
    for(int i = 0; i < paramSlots.size()/5; i++){
		mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent5", "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V");
    }
    
    //Then get the remainder
    switch(paramSlots.size() % 5){
	case 0:
		break;
	case 1:
		mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1", "(Ljava/lang/Object;)V");
		break;
	case 2:
		mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent2", "(Ljava/lang/Object;Ljava/lang/Object;)V");
		break;
	case 3:
		mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent3", "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V");
		break;
	case 4:
		mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent4", "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V");
		break;
    }

  }

  @Override
    protected void onMethodExit(int opcode) {
    if (foundEntry) {
      // System.err.println("onMethodExit: " + className + "::" + methodName);
      // if( (opcode >= Opcodes.IRETURN && opcode <= Opcodes.RETURN) || opcode == Opcodes.ATHROW ){
      
      // -- Emit method ID
      mv.visitLdcInsn(methodId);
      
      if( (access & Opcodes.ACC_STATIC) != 0){
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExitStatic", 
			   "(I)V");
	/* "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V" */
      }
      else{
        /*
	mv.visitVarInsn(Opcodes.ALOAD, 0);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "varEvent1",
			   "(Ljava/lang/Object;)V");
	*/
	mv.visitVarInsn(Opcodes.ALOAD, 0);
	mv.visitMethodInsn(Opcodes.INVOKESTATIC, "ElephantTracks", "methodExit", 
			   "(ILjava/lang/Object;)V");
	/* "(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V" */

      }
      // }
    }
  }

  /*
   * Returns true if fieldDesc is the descriptor fora field that might point to an object.
   */
  private boolean isObjectReference(String fieldDesc){
    if(fieldDesc.startsWith("L") || fieldDesc.startsWith("[")){
      return true;
    }
    else	{
      return false;
    }
  }


}
